

# file utility used to find out mime type of a file
BINARY_FILE = "/usr/bin/file"

# Provided by ImageMagick package.
# Used for resizing images.
BINARY_CONVERT = "/usr/bin/convert"

# Provided by Poppler Utils.
# Used to extract images from PDF file.
BINARY_PDFTOPPM = "/usr/bin/pdftoppm"

# Provided by Poppler Utils.
# used to get page count in PDF file
BINARY_PDFINFO = "/usr/bin/pdfinfo"

# Provided by ImageMagick package.
# Used to get number of pages in TIFF file.
BINARY_IDENTIFY = "/usr/bin/identify"

# Provided by tesseract package.
# Used to extract text from images/PDF files.
BINARY_OCR = "/usr/bin/tesseract"

# Provided by stapler
# Used to edit PDF documents
BINARY_STAPLER = "~/.local/bin/stapler"



import os
import unittest
from mglib.test.utils import TemporaryNode
from mglib.path import DocumentPath
from mglib.storage import FileSystemStorage


BASE_DIR = os.path.dirname(
    os.path.abspath(__file__)
)

MEDIA_ROOT = os.path.join(
    BASE_DIR, "media"
)

class TestStorage(unittest.TestCase):

    def test_delete(self):
        storage = FileSystemStorage(location=MEDIA_ROOT)

        with TemporaryNode(MEDIA_ROOT) as media_root:
            docs = media_root.add_folder("docs")
            res = media_root.add_folder("results")
            f1 = docs.add_folder("user_1/document_2")
            f1.add_file("doku.pdf")
            res.add_folder("user_1/document_2/pages")

            doc_path = DocumentPath(
                user_id=1,
                document_id=2,
                file_name='doku.pdf'
            )

            self.assertTrue(
                f1.exists()
            )

            storage.delete_doc(doc_path)

            self.assertFalse(
                f1.exists()
            )

    def test_get_versions_1(self):
        storage = FileSystemStorage(location=MEDIA_ROOT)

        with TemporaryNode(MEDIA_ROOT) as media_root:
            docs = media_root.add_folder("docs")
            res = media_root.add_folder("results")
            f1 = docs.add_folder("user_1/document_2")
            f1.add_file("doku.pdf")
            # simulate 2 versions of the document.
            f1.add_folder("v1")
            f1.add_folder("v2")
            res.add_folder("user_1/document_2/pages")

            doc_path = DocumentPath(
                user_id=1,
                document_id=2,
                file_name='doku.pdf',
                version=2
            )
            versions = storage.get_versions(doc_path)

            self.assertEqual(
                versions, [0, 1, 2]
            )

    def test_get_versions_2(self):
        storage = FileSystemStorage(location=MEDIA_ROOT)

        with TemporaryNode(MEDIA_ROOT) as media_root:
            docs = media_root.add_folder("docs")
            f1 = docs.add_folder("user_1/document_2")
            f1.add_file("doku.pdf")

            doc_path = DocumentPath(
                user_id=1,
                document_id=2,
                file_name='doku.pdf',
                version=2
            )
            versions = storage.get_versions(doc_path)

            # document has only one version - the latest
            self.assertEqual(
                versions, [0]
            )




import os
import unittest
from unittest import mock
from mglib import stapler
from mglib.conf import settings

BASE_DIR = os.path.dirname(os.path.abspath(__file__))

DATA_DIR = os.path.join(BASE_DIR, "data")


class TestPdfLib(unittest.TestCase):
    def test_ranges_for_reorder(self):
        actual = stapler.cat_ranges_for_reorder(4, [
            {"page_order": 1, "page_num": 4},
            {"page_order": 2, "page_num": 3},
            {"page_order": 3, "page_num": 2},
            {"page_order": 4, "page_num": 1}
        ])
        expected = [4, 3, 2, 1]
        assert expected == actual

        self.assertRaises(ValueError, stapler.cat_ranges_for_reorder, 2, [])
        self.assertRaises(KeyError, stapler.cat_ranges_for_reorder, 2, [
            {"page_order": 3, "page_num": 4},
            {"page_order": 5, "page_num": 6}
        ])

    def test_delete_pages(self):
        input_file = os.path.join(DATA_DIR, "berlin.pdf")
        output_file = os.path.join(DATA_DIR, "berlin2.pdf")

        with mock.patch("mglib.stapler.run") as run_func:
            stapler.delete_pages(input_file, output_file, [1])
            run_func.assert_called()
            run_func.assert_called_with(
                [settings.BINARY_STAPLER, "del", input_file, "1", output_file]
            )

    def test_split_ranges(self):
        page_count = 9

        self.assertRaises(
            ValueError,
            stapler.split_ranges,
            9,
            after="a",
            before=False
        )
        self.assertRaises(
            ValueError,
            stapler.split_ranges,
            9, after=False,
            before=True
        )

        actual1, actual2 = stapler.split_ranges(page_count, 1, False)
        expected1 = [1]
        expected2 = [2, 3, 4, 5, 6, 7, 8, 9]
        assert actual1 == expected1
        assert actual2 == expected2

        actual1, actual2 = stapler.split_ranges(page_count, False, 2)
        expected1 = [1]
        expected2 = [2, 3, 4, 5, 6, 7, 8, 9]
        assert actual1 == expected1
        assert actual2 == expected2

        actual1, actual2 = stapler.split_ranges(page_count)
        expected1 = list(range(1, page_count + 1))
        expected2 = []
        assert actual1 == expected1
        assert actual2 == expected2

    def test_reorder_pages(self):
        input_file = os.path.join(DATA_DIR, "berlin.pdf")
        output_file = os.path.join(DATA_DIR, "berlin2.pdf")
        new_order = [
            {'page_num': 2, 'page_order': 1},
            {'page_num': 1, 'page_order': 2},
        ]

        with mock.patch("mglib.stapler.run") as run_func:
            stapler.reorder_pages(input_file, output_file, new_order)
            run_func.assert_called()
            run_func.assert_called_with(
                [
                    settings.BINARY_STAPLER,
                    "sel",
                    input_file,
                    "2",
                    "1",
                    output_file
                ]
            )

    def test_paste_pages_into_existing_doc(self):
        input_file = os.path.join(DATA_DIR, "berlin.pdf")
        output_file = os.path.join(DATA_DIR, "berlin2.pdf")
        datalist = []

        with mock.patch("mglib.stapler.run") as run_func:
            stapler.paste_pages_into_existing_doc(
                input_file, output_file, datalist
            )
            run_func.assert_called()
            run_func.assert_called_with(
                [
                    settings.BINARY_STAPLER,
                    "sel", "A=" + input_file, "A1", "A2", output_file
                ]
            )

        datalist = [{"src": input_file, "page_nums": "34"}]

        with mock.patch("mglib.stapler.run") as run_func:
            stapler.paste_pages_into_existing_doc(
                input_file,
                output_file,
                datalist,
                1
            )
            run_func.assert_called()
            run_func.assert_called_with(
                [
                    settings.BINARY_STAPLER,
                    "sel", "A=" + input_file,
                    "B=" + input_file, "A1", "B3",
                    "B4", "A2", output_file
                ]
            )

    def test_paste_pages(self):
        input_file = os.path.join(DATA_DIR, "berlin.pdf")
        output_file = os.path.join(DATA_DIR, "berlin2.pdf")
        datalist = []

        with mock.patch("mglib.stapler.run") as run_func:
            stapler.paste_pages(input_file, output_file, datalist, False)
            run_func.assert_called()
            run_func.assert_called_with(
                [
                    settings.BINARY_STAPLER,
                    "sel",
                    "A=" + input_file,
                    "A1",
                    "A2",
                    output_file
                ]
            )

        datalist = [{"src": input_file, "page_nums": "34"}]

        with mock.patch("mglib.stapler.run") as run_func:
            stapler.paste_pages(input_file, output_file, datalist)
            run_func.assert_called()
            run_func.assert_called_with(
                [settings.BINARY_STAPLER, "sel", "A=" + input_file, "A3", "A4",
                    output_file]
            )




import os
from pathlib import Path
import unittest

from mglib.conf.settings import (
    MgLibSettings,
    DefaultSettings
)

DATA_DIR = os.path.join(
    Path(__file__).parent,
    'data'
)


class TestMgLibSettings(unittest.TestCase):

    def setUp(self):
        self.settings = MgLibSettings(DefaultSettings())

    def test_settings_outside_django_should_work(self):
        """
        Without django there should be default values
        for settings
        """
        # check default value for pdfinfo
        self.assertEqual(
            "/usr/bin/pdfinfo",
            self.settings.BINARY_PDFINFO
        )

    def test_settings_are_configurable(self):
        """
        User should be able to reconfigure mglibsettings
        on the go (i.e. change default values).
        """
        # check default value for pdfinfo
        self.settings.configure(
            BINARY_PDFINFO="/usr/bin/xyz"
        )
        self.assertEqual(
            "/usr/bin/xyz",
            self.settings.BINARY_PDFINFO
        )




import importlib

try:
    from django.conf import settings as django_settings
except ImportError:
    # Operating outside django, use own settings module
    django_settings = None


class DefaultSettings:

    def __init__(
        self,
        settings_module="mglib.conf.default_settings"
    ):
        self.SETTINGS_MODULE = settings_module

        mod = importlib.import_module(
            self.SETTINGS_MODULE
        )

        for setting in dir(mod):
            if setting.isupper():
                setting_value = getattr(mod, setting)
                setattr(self, setting, setting_value)

    def configure(self, **options):
        for name, value in options.items():
            setattr(self, name, value)


class MgLibSettings:

    def __init__(
        self, default_settings
    ):
        self.default_settings = default_settings

    def __getattr__(self, name):
        # When operating withing django,
        # get configuration from django settings
        if not name.isupper():
            raise AttributeError

        if django_settings:
            val = getattr(django_settings, name)
            return val

        val = getattr(self.default_settings, name)
        return val

    def configure(self, **options):
        self.default_settings.configure(
            **options
        )



import os
import re
import subprocess
import logging
from magic import from_file

from .conf import settings
from .exceptions import FileTypeNotSupported

"""
Uses command line pdfinfo utility (from poppler pakage) for various
small operations (e.g. get pdf page count).
"""

logger = logging.getLogger(__name__)


def get_tiff_pagecount(filepath):
    cmd = [
        settings.BINARY_IDENTIFY,
        "-format",
        "%n\n",
        filepath
    ]
    compl = subprocess.run(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE
    )

    if compl.returncode:

        logger.error(
            "get_tiff_pagecount: cmd=%s args=%s stdout=%s stderr=%s code=%s",
            cmd,
            compl.args,
            compl.stdout,
            compl.stderr,
            compl.returncode,
            stack_info=True
        )

        raise Exception("Error occured while getting document page count.")

    lines = _split(stdout=compl.stdout)
    # look up for the line containing "Pages: 11"
    for line in lines:
        x = re.match(r"(\d+)", line.strip())
        if x:
            return int(x.group(1))

    return 0


def get_pagecount(filepath):
    """
    Returns the number of pages in a PDF document as integer.

    filepath - is filesystem path to a PDF document
    """
    if not os.path.isfile(filepath):
        raise ValueError("Filepath %s is not a file" % filepath)

    if os.path.isdir(filepath):
        raise ValueError("Filepath %s is a directory!" % filepath)

    base, ext = os.path.splitext(filepath)
    mime_type = from_file(filepath, mime=True)
    # pure images (png, jpeg) have only one page :)

    if mime_type in ['image/png', 'image/jpeg', 'image/jpg']:
        # whatever png/jpg image is there - it is
        # considered by default one page document.
        return 1

    # In case of REST API upload (via PUT + form multipart)
    # django saves temporary file as application/octet-stream
    # Checking extentions is an extra method of finding out correct
    # mime type
    if ext and ext.lower() in ('.jpeg', '.png', '.jpg'):
        return 1

    if mime_type == 'image/tiff':
        return get_tiff_pagecount(filepath)

    # In case of REST API upload (via PUT + form multipart)
    # django saves temporary file as application/octet-stream
    # Checking extentions is an extra method of finding out correct
    # mime type
    if ext and ext.lower() in ('.tiff', ):
        return get_tiff_pagecount(filepath)

    if mime_type != 'application/pdf':
        # In case of REST API upload (via PUT + form multipart)
        # django saves temporary file as application/octet-stream
        # Checking extentions is an extra method of finding out correct
        # mime type
        if ext and ext.lower() != '.pdf':
            raise FileTypeNotSupported(
                "Only jpeg, png, pdf and tiff are handled by this"
                " method"
            )
    # pdfinfo "${PDFFILE}" | grep Pages
    cmd = [
        settings.BINARY_PDFINFO,
        filepath
    ]
    compl = subprocess.run(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE
    )

    if compl.returncode:

        logger.error(
            "get_pagecount: cmd=%s args=%s stdout=%s stderr=%s code=%s",
            cmd,
            compl.args,
            compl.stdout,
            compl.stderr,
            compl.returncode,
            stack_info=True
        )

        raise Exception("Error occured while getting document page count.")

    lines = _split(stdout=compl.stdout)
    # look up for the line containing "Pages: 11"
    for line in lines:
        x = re.match(r"Pages:\W+(\d+)$", line.strip())
        if x:
            return int(x.group(1))

    return 0


def _split(stdout):
    """
    stdout is result.stdout where result
    is whatever is returned by subprocess.run
    """
    decoded_text = stdout.decode(
        'utf-8',
        # in case there are decoding issues, just replace
        # problematic characters. We don't need text verbatim.
        'replace'
    )
    lines = decoded_text.split('\n')

    return lines



import logging
from . import wrapper

from .conf import settings

logger = logging.getLogger(__name__)


class Mime(wrapper.Wrapper):
    def __init__(self, filepath):
        super().__init__(exec_name=settings.BINARY_FILE)
        self.filepath = filepath

    def get_cmd(self):
        cmd = super().get_cmd()

        cmd.extend(['--mime-type'])
        cmd.extend(['-b'])
        cmd.extend([self.filepath])

        return cmd

    def is_tiff(self):
        return self.guess() == 'image/tiff'

    def is_pdf(self):
        return self.guess() == 'application/pdf'

    def is_image(self):
        """
        Returns true if MIME type is one of following:
            * image/png
            * image/jpg
        """
        return self.guess() in ('image/png', 'image/jpg', 'image/jpeg')

    def guess(self):
        cmd = self.get_cmd()
        complete = self.run(cmd)

        return complete.stdout.strip()

    def __str__(self):

        mime_type = self.guess()
        return f"Mime({self.filepath}, {mime_type})"



import logging
import subprocess

logger = logging.getLogger(__name__)


class Wrapper:

    def __init__(self, exec_name, check=True, dry_run=False):
        self.exec_name = exec_name
        self.check = check,
        # usefull for debugging purpose
        self.dry_run = dry_run

    def get_cmd(self):
        cmd = []

        if self.exec_name:
            cmd.extend([self.exec_name])

        return cmd

    def call_no_args(self):

        cmd = self.get_cmd()
        self.run(cmd)

    def run(self, cmd):

        command_to_run = ' '.join(cmd)

        if (self.dry_run):
            logger.debug(f"Dry run: {command_to_run}")

        logger.debug(f"subprocess: {command_to_run}")

        ret = subprocess.run(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            encoding="utf-8"
        )
        if ret.returncode != 0:
            logger.error((
                f"returncode={ret.returncode}"
                f" stdout={ret.stdout}"
                f" stderr={ret.stderr}"
            ))
        return ret



import unittest

from mglib.path import (
    DocumentPath, PagePath,
)
from mglib.step import Step


class TestDocumentPath(unittest.TestCase):

    def test_document_url(self):
        doc_ep = DocumentPath(
            user_id=1,
            document_id=3,
            file_name="x.pdf"
        )
        self.assertEqual(
            doc_ep.url(),
            "docs/user_1/document_3/x.pdf"
        )

    def test_document_url_with_another_version(self):

        doc_ep = DocumentPath(
            user_id=1,
            document_id=15,
            file_name="x.pdf"
        )
        self.assertEqual(
            doc_ep.url(version=3),
            "docs/user_1/document_15/v3/x.pdf"
        )

        self.assertEqual(
            doc_ep.url(version=2),
            "docs/user_1/document_15/v2/x.pdf"
        )

    def test_document_url_none_vs_0(self):
        doc_ep = DocumentPath(
            user_id=1,
            document_id=15,
            file_name="x.pdf"
        )
        doc_ep.inc_version()  # current version = 1
        doc_ep.inc_version()  # current version = 2
        doc_ep.inc_version()  # current version = 3

        self.assertEqual(
            # with version == None, latest version of the document
            # will be returned, which is 3
            doc_ep.url(version=None),
            "docs/user_1/document_15/v3/x.pdf"
        )

        self.assertEqual(
            # with version == 0, version 0 will be provided
            # i.e. version=0 returns original doc.
            doc_ep.url(version=0),
            "docs/user_1/document_15/x.pdf"
        )

    def test_inc_version(self):
        """
        Document endpoints are now versioned.
        Initial version is 0.
        When version is 0, the "old" endpoint path applies i.e.
        version is not included in the path.
        After document is modified (blank page deleted for example),
        its version is incremented. If document version is > 0, then
        version is included in the path.
        """
        doc_ep = DocumentPath(
            user_id=1,
            document_id=3,
            file_name="x.pdf"
        )
        doc_ep.inc_version()

        self.assertEqual(
            doc_ep.url(),
            "docs/user_1/document_3/v1/x.pdf"
        )

        doc_ep.inc_version()

        self.assertEqual(
            doc_ep.url(),
            "docs/user_1/document_3/v2/x.pdf"
        )

        # however, explicit version can be forced
        # by providing an argument to url method.
        self.assertEqual(
            doc_ep.url(version=1),
            "docs/user_1/document_3/v1/x.pdf"
        )

    def test_dirname(self):
        ep = DocumentPath(
            user_id=1,
            document_id=3,
            aux_dir="results",
            file_name="x.pdf"
        )
        self.assertEqual(
            ep.dirname(),
            "results/user_1/document_3/"
        )

    def test_pages_dirname(self):
        ep = DocumentPath(
            user_id=1,
            document_id=3,
            aux_dir="results",
            file_name="x.pdf"
        )
        self.assertEqual(
            ep.pages_dirname(),
            "results/user_1/document_3/pages/"
        )


class TestPagePath(unittest.TestCase):

    def test_versioned_page_ep(self):
        doc_path = DocumentPath(
            user_id=1,
            document_id=3,
            file_name="x.pdf"
        )
        # document's version incremented
        doc_path.inc_version()

        page_ep = PagePath(
            document_path=doc_path,
            page_num=1,
            page_count=3
        )
        self.assertEqual(
            page_ep.path,
            "results/user_1/document_3/v1/pages/page_1.txt"
        )

    def test_txt_url(self):
        """
        Without any arguments
            page_ep.url() returns page_ep.txt_url()
        """
        doc_ep = DocumentPath(
            user_id=1,
            document_id=3,
            file_name="x.pdf"
        )
        page_ep = PagePath(
            document_path=doc_ep,
            page_num=1,
            step=Step(1),
            page_count=3
        )
        self.assertEqual(
            page_ep.url(),
            page_ep.txt_url()
        )

    def test_ppmroot(self):
        doc_ep = DocumentPath(
            user_id=1,
            document_id=3,
            file_name="x.pdf"
        )
        page_url = PagePath(
            document_path=doc_ep,
            page_num=1,
            step=Step(1),
            page_count=3
        )
        self.assertEqual(
            page_url.ppmroot,
            "results/user_1/document_3/pages/page_1/100/page"
        )

#    def test_hocr_exists(self):
#        local_media = os.path.join(
#            os.path.dirname(os.path.dirname(__file__)),
#            "test",
#            "media"
#        )
#        remote_ep = Endpoint("s3:/test-papermerge/")
#        local_ep = Endpoint(f"local:{local_media}")
#        doc_ep = DocumentPath(
#            remote_endpoint=remote_ep,
#            local_endpoint=local_ep,
#            user_id=1,
#            document_id=3,
#            file_name="x.pdf"
#        )
#        page_ep1 = PagePath(
#            document_ep=doc_ep,
#            page_num=1,
#            step=Step(1),
#            page_count=3
#        )
#        self.assertTrue(
#            page_ep1.hocr_exists()
#        )
#        page_ep2 = PagePath(
#            document_ep=doc_ep,
#            page_num=2,
#            step=Step(1),
#            page_count=3
#        )
#        self.assertFalse(
#            page_ep2.hocr_exists()
#        )





import os
import logging

from mglib.runcmd import run
from .conf import settings

logger = logging.getLogger(__name__)


def pdfname_from_tiffname(doc_url):
    """
    Given tiff document url, will return
    respective pdf file name. Returned
    file name can be use used as destination
    for tiff2pdf tool.

    Returns a tuple (new_doc_url, new_filename).
    new_doc_url - is new absolute path to the pdf file
    new_filename - is new pdf filename
    """
    # basename is filename + ext (no path)
    basename = os.path.basename(doc_url)
    base_root, base_ext = os.path.splitext(basename)
    root, ext = os.path.splitext(doc_url)
    new_doc_url = f"{root}.pdf"

    return new_doc_url, f"{base_root}.pdf"


def convert_tiff2pdf(doc_url):

    logger.debug(f"convert_tiff2pdf for {doc_url}")

    new_doc_url, new_filename = pdfname_from_tiffname(
        doc_url
    )

    logger.debug(
        f"tiff2pdf source={doc_url} dest={new_doc_url}"
    )

    cmd = (
        settings.BINARY_CONVERT,
        doc_url,
        new_doc_url,
    )

    run(cmd)

    # returns new filename
    return new_filename



import os
import logging

from mglib.runcmd import run
from .conf import settings

logger = logging.getLogger(__name__)


def resize_img(page_path, media_root):

    local_abspath = os.path.join(
        media_root,
        page_path.document_path.url()
    )
    logger.debug(f"Resizing image {page_path.img_url()}")

    ppmroot = os.path.join(media_root, page_path.ppmroot)
    ppmroot_dirname = os.path.dirname(ppmroot)

    width = page_path.step.width

    if not os.path.exists(ppmroot_dirname):
        logger.debug(f"PPMROOT {ppmroot_dirname} does not exists. Creating.")
        os.makedirs(
            ppmroot_dirname, exist_ok=True
        )
    else:
        logger.debug(f"PPMROOT {ppmroot_dirname} already exists.")

    cmd = (
        settings.BINARY_CONVERT,
        "-resize",
        f"{width}x",
        local_abspath,
        # output directory path, similar to ppmroot
        f"{ppmroot}-1.jpg"
    )

    run(cmd)


def extract_img(page_path, media_root):

    local_abspath = os.path.join(
        media_root,
        page_path.document_path.url()
    )
    logger.debug(f"Extracing image for {page_path.img_url()}")

    ppmroot = os.path.join(media_root, page_path.ppmroot)
    ppmroot_dirname = os.path.dirname(ppmroot)

    page_num = page_path.page_num
    width = page_path.step.width

    if not os.path.exists(ppmroot_dirname):
        logger.debug(f"PPMROOT {ppmroot_dirname} does not exists. Creating.")
        os.makedirs(
            ppmroot_dirname, exist_ok=True
        )
    else:
        logger.debug(f"PPMROOT {ppmroot_dirname} already exists.")
    cmd = (
        settings.BINARY_PDFTOPPM,
        "-jpeg",
        "-f",
        str(page_num),
        "-l",  # generate only one page
        str(page_num),
        "-scale-to-x",
        str(width),
        "-scale-to-y",
        "-1",  # it will adjust height according to img ratio
        local_abspath,
        # output directory path,
        ppmroot
    )

    run(cmd)


def extract_hocr(page_url, lang, media_root):
    page_abspath = os.path.join(
        media_root,
        page_url.img_url()
    )

    hocr_root, hocr_ext = os.path.splitext(
        os.path.join(media_root, page_url.hocr_url())
    )
    cmd = (
        settings.BINARY_OCR,
        "-l",
        lang,
        page_abspath,
        hocr_root,
        "hocr"
    )
    run(cmd)
    logger.debug(f"OCR for {page_url.img_url()} - Complete.")
    logger.debug(f"OCR Result {page_url.hocr_url()}.")


def extract_txt(page_url, lang, media_root):
    page_abspath = os.path.join(
        media_root,
        page_url.img_url()
    )
    txt_root, txt_ext = os.path.splitext(
        os.path.join(
            media_root, page_url.txt_url()
        )
    )
    cmd = (
        settings.BINARY_OCR,
        "-l",
        lang,
        page_abspath,
        txt_root
    )
    run(cmd)



import logging

from mglib.runcmd import run
from mglib.pdfinfo import get_pagecount

from .conf import settings

logger = logging.getLogger(__name__)

#
#  Utilities around stapler command line tool
#
#  https://github.com/hellerbarde/stapler
#


def cat_ranges_for_reorder(page_count, new_order):
    """
    Returns a list of integers. Each number in the list
    is correctly positioned (newly ordered) page.

    Examples:

    If in document with 4 pages first and second pages were
    swapped, then returned list will be:

        [2, 1, 3, 4]

    If first page was swapped with last one (also 4 paegs document)
    result list will look like:

        [4, 2, 3, 1]
    """
    if len(new_order) != page_count:
        raise ValueError("Not enough pages specified")
    results = []
    # key = page_num
    # value = page_order
    page_map = {}

    for item in new_order:
        k = int(item['page_order'])
        v = int(item['page_num'])
        page_map[k] = v

    for number in range(1, page_count + 1):
        results.append(
            page_map[number]
        )

    return results


def split_ranges(total, after=False, before=False):
    """
    Given a range 1, 2, ..., total (page numbers of a doc).
    Split it in two lists.
    Example:
    Input: total = 9, after=1, before=False
    Output: list1 = [1]; list2 = [2, 3, 4, ..., 9].

    Input: total = 9; after=False, before=1
    Output: list1 = [], list2 = [1, 2, 3, 4, ..., 9]

    Input: total = 5; after=4; before=False
    Output: list1 = [1, 2, 3, 4] list2 = [5]

    Input: total = 5; after=False; before=False;
    Output: list1 = [1, 2, 3, 4, 5], list2 = []
    (it means, by default, all pages are inserted at the end of the doc)
    """
    if after and not before:
        if not type(after) == int:
            raise ValueError(
                "argument 'after' is supposed to be an int"
            )
        list1 = list(range(1, after + 1))
        list2 = list(range(after + 1, total + 1))
        return list1, list2

    if not after and before:
        if not type(before) == int:
            raise ValueError(
                "argument 'before' is supposed to be an int"
            )
        list1 = list(range(1, before))
        list2 = list(range(before, total + 1))
        return list1, list2

    list1 = list(range(1, total + 1))
    list2 = []

    return list1, list2


def paste_pages_into_existing_doc(
    src,
    dst,
    data_list,
    after_page_number=False,
    before_page_number=False
):
    page_count = get_pagecount(src)
    list1, list2 = split_ranges(
        total=page_count,
        after=after_page_number,
        before=before_page_number
    )
    # notice missing A
    # Letter A is assignent to current folder and
    # pages from list1 and list2
    letters = "BCDEFGHIJKLMNOPQRSTUVWXYZ"
    letters_2_doc_map = []
    letters_pages = []
    letters_pages_before = []
    letters_pages_after = []

    letters_2_doc_map.append(
        f"A={src}"
    )

    for idx in range(0, len(data_list)):
        letter = letters[idx]
        src = data_list[idx]['src']
        pages = data_list[idx]['page_nums']

        letters_2_doc_map.append(
            f"{letter}={src}"
        )
        for p in pages:
            letters_pages.append(
                f"{letter}{p}"
            )

    for p in list1:
        letters_pages_before.append(
            f"A{p}"
        )

    for p in list2:
        letters_pages_after.append(
            f"A{p}"
        )

    cmd = [
        settings.BINARY_STAPLER,
    ]
    cmd.append("sel")

    # add A=doc1_path, B=doc2_path
    cmd.extend(letters_2_doc_map)

    # existing doc pages (may be empty)
    cmd.extend(letters_pages_before)
    # newly inserted pages
    cmd.extend(letters_pages)
    # existing doc pages (may be empty)
    cmd.extend(letters_pages_after)

    cmd.append(dst)

    run(cmd)


def paste_pages(
    src,
    dst,
    data_list,
    dst_doc_is_new=True,
    after_page_number=False,
    before_page_number=False
):
    """
    dest_doc_ep = endpoint of the doc where newly created
        file will be placed.
    src_doc_ep_list is a list of following format:
        [
            {
                'doc_ep': doc_ep,
                'page_nums': [page_num_1, page_num_2, page_num_3]
            },
            {
                'doc_ep': doc_ep,
                'page_nums': [page_num_1, page_num_2, page_num_3]
            },
            ...
        ]
    src_doc_ep_list is a list of documents where pages
    (with numbers page_num_1...) will be paste from.

    dst_doc_is_new = True well.. destination document was just created,
    we are pasting here cutted pages into some folder as new document.

    In this case 'after' and 'before' arguments are ignored

    dst_doc_is_new = False, pasting pages into exiting document.
    If before_page_number > 0 - paste pages before page number
        'before_page_number'
    If after_page_number > 0 - paste pages after page number
        'after_page_number'

    before_page_number argument has priority over after_page_number.

    If both before_page_number and after_page_number are < 0 - just paste
    pages at the end of the document.
    """
    if not dst_doc_is_new:
        return paste_pages_into_existing_doc(
            src=src,
            dst=dst,
            data_list=data_list,
            after_page_number=after_page_number,
            before_page_number=before_page_number
        )
    letters = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    letters_2_doc_map = []
    letters_pages = []

    for idx in range(0, len(data_list)):
        letter = letters[idx]
        src = data_list[idx]['src']
        pages = data_list[idx]['page_nums']

        letters_2_doc_map.append(
            f"{letter}={src}"
        )
        for p in pages:
            letters_pages.append(
                f"{letter}{p}"
            )

    cmd = [
        settings.BINARY_STAPLER,
    ]
    cmd.append("sel")

    # add A=doc1_path, B=doc2_path
    cmd.extend(letters_2_doc_map)

    cmd.extend(letters_pages)

    cmd.append(dst)

    run(cmd)


def reorder_pages(
    src, dst, new_order
):
    """
    new_order is a list of following format:

        [
            {'page_num': 2, page_order: 1},
            {'page_num': 1, page_order: 2},
            {'page_num': 3, page_order: 3},
            {'page_num': 4, page_order: 4},
        ]
    Example above means that in current document of 4 pages,
    first page was swapped with second one.
    page_num    = older page order
    page_order  = current page order
    So in human language, each hash is read:
        <page_num> now should be <page_order>
    """
    page_count = get_pagecount(src)

    cat_ranges = cat_ranges_for_reorder(
        page_count=page_count,
        new_order=new_order
    )

    cmd = [
        settings.BINARY_STAPLER,
        "sel",
        src
    ]
    for page in cat_ranges:
        cmd.append(
            str(page)
        )

    cmd.append(dst)
    run(cmd)


def delete_pages(src, dst, page_numbers):
    cmd = [
        settings.BINARY_STAPLER,
        "del",
        src
    ]
    for page in page_numbers:
        cmd.append(
            str(page)
        )

    cmd.append(dst)

    run(cmd)



import logging

logger = logging.getLogger(__name__)

AUX_DIR_DOCS = "docs"
AUX_DIR_RESULTS = "results"


class DocumentPath:
    """
    Document path:
    /<aux_dir>/<user_id>/<doc_id>/<version>/<file_name>

    If version = 0, it is not included in DocumentPath.
    Document's version is incremented everytime pdftk operation runs on it
    (when pages are deleted, reordered, pasted)
    """

    def __init__(
        self,
        user_id,
        document_id,
        file_name,
        aux_dir=AUX_DIR_DOCS,
        version=0
    ):
        self.user_id = user_id
        self.document_id = document_id
        self.file_name = file_name
        self.aux_dir = aux_dir
        # by default, document has version 0
        self.version = version
        self.pages = "pages"

    def url(self, version=None):
        if version:
            version = int(version)

        return f"{self.dirname(version=version)}{self.file_name}"

    def path(self, version=None):
        if version:
            version = int(version)
        return self.url(version=version)

    @property
    def dirname_docs(self):
        _path = (
            f"{AUX_DIR_DOCS}/user_{self.user_id}/"
            f"document_{self.document_id}/"
        )

        return _path

    @property
    def dirname_results(self):
        _path = (
            f"{AUX_DIR_RESULTS}/user_{self.user_id}/"
            f"document_{self.document_id}/"
        )

        return _path

    def dirname(self, version=None):

        if version is None:
            version = self.version

        full_path = (
            f"{self.aux_dir}/user_{self.user_id}/"
            f"document_{self.document_id}/"
        )

        if version > 0:
            full_path = f"{full_path}v{version}/"

        return full_path

    def pages_dirname(self, version=None):
        return f"{self.dirname(version=version)}{self.pages}/"

    def __repr__(self):
        message = (
            f"DocumentPath(version={self.version},"
            f"user_id={self.user_id},"
            f"document_id={self.document_id},"
            f"file_name={self.file_name})"
        )
        return message

    def inc_version(self):
        self.version = self.version + 1

    def copy_from(doc_path, **kw):
        """
        Will create a copy of provided
        DocumentPath (first parameter = doc_path) and replace
        existing parameter of new copy with the one from kw.

        kw => key/value parameters.
        Keys can be one of doc_path attributes: user_id, document_id,
        file_name, aux_dir, version
        """
        copy_values = {
            'user_id': doc_path.user_id,
            'document_id': doc_path.document_id,
            'file_name': doc_path.file_name,
            'version': doc_path.version,
            'aux_dir': doc_path.aux_dir

        }
        for key, value in kw.items():
            copy_values[key] = kw[key]

        return DocumentPath(
            user_id=copy_values['user_id'],
            document_id=copy_values['document_id'],
            file_name=copy_values['file_name'],
            version=copy_values['version'],
            aux_dir=copy_values['aux_dir']
        )


class PagePath:
    """
    <aux_dir>/<doc_id>/pages/<page_num>/<step>/page-<xyz>.jpg
    """

    def __init__(
        self,
        document_path,
        page_num,
        page_count,
        step=None
    ):
        if not isinstance(page_num, int):
            msg_err = f"PagePath.page_num must be an int. Got {page_num}."
            raise ValueError(msg_err)

        self.document_path = document_path
        self.results_document_ep = DocumentPath.copy_from(
            document_path,
            aux_dir=AUX_DIR_RESULTS
        )
        self.page_count = page_count
        self.page_num = page_num
        self.step = step
        self.pages = self.document_path.pages

    @property
    def ppmroot(self):
        # returns schema://.../<doc_id>/pages/<page_num>/<step>/page
        pages_dirname = self.results_document_ep.pages_dirname()
        result = (
            f"{pages_dirname}page_{self.page_num}/"
            f"{self.step.percent}/page"
        )
        return result

    @property
    def pages_dirname(self):
        return self.document_path.pages_dirname()

    @property
    def path(self):
        return self.url()

    def url(self):
        return self.txt_url()

    @property
    def txt_path(self):
        return self.txt_url()

    def txt_url(self):
        pages_dirname = self.results_document_ep.pages_dirname()
        return f"{pages_dirname}page_{self.page_num}.txt"

    @property
    def hocr_path(self):
        return self.hocr_url()

    def hocr_url(self):
        url = f"{self.ppmroot}-{self.ppmtopdf_formated_number}.hocr"
        return url

    @property
    def img_path(self):
        return self.img_url()

    def img_url(self):
        url = f"{self.ppmroot}-{self.ppmtopdf_formated_number}.jpg"
        return url

    @property
    def ppmtopdf_formated_number(self):

        if self.page_count <= 9:
            fmt_num = "{num:d}"
        elif self.page_count > 9 and self.page_count < 100:
            fmt_num = "{num:02d}"
        elif self.page_count >= 100:
            fmt_num = "{num:003d}"

        return fmt_num.format(
            num=int(self.page_num)
        )



import logging
import os
import shutil
from os import listdir
from os.path import isdir, join

from mglib import stapler
from mglib.path import DocumentPath, PagePath
from mglib.step import Steps
from mglib.utils import get_assigns_after_delete, safe_to_delete

logger = logging.getLogger(__name__)


class Storage:
    """
    Default Storage class which works with DocumentPath and PagePath
    on local host filesystem
    """

    def __init__(self, location=None, **kwargs):
        # by default, this will be something like
        # settings.MEDIA_ROOT
        self._location = location

    @property
    def location(self):
        return self._location

    def upload(self, doc_path_url, **kwargs):
        pass

    def download(self, doc_path_url, **kwargs):
        pass

    def make_sure_path_exists(self, filepath):
        logger.debug(f"make_sure_path_exists {filepath}")
        dirname = os.path.dirname(filepath)
        os.makedirs(
            dirname,
            exist_ok=True
        )

    def get_versions(self, doc_path):
        """
        Returns a list of (all) ordered versions
        of specific doc_path. Versions
        start with 0. Examples of return values:

        - [0, 1, 2, 3] = 4 versions of the document
        - [ 0 ] = only one version (original)

        To count versions it just counts number of subfolders
        in specific document folder. Versions are
        stored in subfolders named v1, v2, v3, ...
        """
        abs_dirname_docs = self.path(
            doc_path.dirname_docs
        )
        try:
            only_dirs = [
                fi for fi in listdir(abs_dirname_docs) if isdir(
                    join(abs_dirname_docs, fi)
                )
            ]
        except FileNotFoundError:
            # in tests, document folders are not always created.
            # If no document folder is found, just return [ 0 ]
            # i.e that document has only one single version and it
            # is the latest one.
            return [0]

        dirs_count = len(only_dirs)

        return list(range(0, dirs_count + 1))

    def get_pagecount(self, doc_path):
        """
        Returns total number of pages for this doc_path.
        Total number of pages = number of page_xy.txt files
        in pages_dirname folder.
        """
        doc_path_pointing_to_results = DocumentPath.copy_from(
            doc_path, aux_dir="results"
        )
        pages_dir = self.abspath(doc_path_pointing_to_results.pages_dirname())

        only_dirs = [
            fi for fi in listdir(pages_dir) if isdir(join(pages_dir, fi))
        ]
        return len(only_dirs)

    def abspath(self, _path):
        if isinstance(_path, DocumentPath):
            return os.path.join(
                self.location, _path.url()
            )
        elif isinstance(_path, PagePath):
            return os.path.join(
                self.location, _path.url()
            )

        return os.path.join(
            self.location, _path
        )

    def path(self, _path):
        return self.abspath(_path)

    def delete_doc(self, doc_path):
        """
        Receives a mglib.path.DocumentPath instance
        """
        # where original documents and their versions are stored
        abs_dirname_docs = self.path(
            doc_path.dirname_docs
        )
        # where OCRed information and generated thumbnails
        # are stored
        abs_dirname_results = self.path(
            doc_path.dirname_results
        )
        # Before recursively deleting everything in folder
        # double check that there are only
        # .pdf, .txt, .hocr, .jpg files.
        if safe_to_delete(
            abs_dirname_docs
        ):
            shutil.rmtree(abs_dirname_docs)
            if os.path.exists(abs_dirname_docs):
                os.rmdir(abs_dirname_docs)

        if safe_to_delete(
            abs_dirname_results
        ):
            shutil.rmtree(abs_dirname_results)
            if os.path.exists(abs_dirname_results):
                os.rmdir(abs_dirname_results)

    def copy_doc(self, src: DocumentPath, dst: DocumentPath):
        """
        copy given file src file path to destination
        as absolute doc_path
        """

        dirname = os.path.dirname(
            self.abspath(dst)
        )
        if not os.path.exists(
            dirname
        ):
            os.makedirs(
                dirname, exist_ok=True
            )
        logger.debug(
            f"copy_doc: {src} to {dst}"
        )
        shutil.copyfile(
            self.abspath(src),
            self.abspath(dst)
        )

    def exists(self, _path):
        return os.path.exists(
            self.path(_path)
        )

    def copy_page_txt(self, src_page_path, dst_page_path):

        self.make_sure_path_exists(
            self.abspath(dst_page_path.txt_url())
        )

        src_txt = self.abspath(src_page_path.txt_url())
        dst_txt = self.abspath(dst_page_path.txt_url())

        logger.debug(f"copy src_txt={src_txt} dst_txt={dst_txt}")
        shutil.copy(src_txt, dst_txt)

    def copy_page_img(self, src_page_path, dst_page_path):

        self.make_sure_path_exists(
            self.abspath(dst_page_path.img_url())
        )

        src_img = self.abspath(src_page_path.img_url())
        dst_img = self.abspath(dst_page_path.img_url())
        logger.debug(f"copy src_img={src_img} dst_img={dst_img}")
        shutil.copy(src_img, dst_img)

    def copy_page_hocr(self, src_page_path, dst_page_path):

        self.make_sure_path_exists(
            self.abspath(dst_page_path.hocr_url())
        )

        src_hocr = self.abspath(src_page_path.hocr_url())
        dst_hocr = self.abspath(dst_page_path.hocr_url())
        logger.debug(f"copy src_hocr={src_hocr} dst_hocr={dst_hocr}")
        shutil.copy(src_hocr, dst_hocr)

    def copy_page(self, src_page_path, dst_page_path):
        """
        Copies page data from source to destination.

        Page data are files with following extentions:
            * txt
            * hocr
            * jpeg
        they are located in media root of respective application.
        """
        for inst in [src_page_path, dst_page_path]:
            if not isinstance(inst, PagePath):
                raise ValueError("copy_page accepts only PagePath instances")

        # copy .txt file
        if self.exists(src_page_path.txt_url()):
            self.copy_page_txt(
                src_page_path=src_page_path,
                dst_page_path=dst_page_path
            )
        else:
            logger.debug(
                f"txt does not exits {src_page_path.txt_url()}"
            )

        # hocr
        if self.exists(src_page_path.hocr_url()):
            self.copy_page_hocr(
                src_page_path=src_page_path,
                dst_page_path=dst_page_path
            )
        else:
            logger.debug(
                f"hocr does not exits {src_page_path.hocr_url()}"
            )

        if src_page_path.img_url():
            self.copy_page_img(
                src_page_path=src_page_path,
                dst_page_path=dst_page_path
            )
        else:
            logger.debug(
                f"img does not exits {src_page_path.img_url()}"
            )

    def reorder_pages(self, doc_path, new_order):
        """
        Reorders pages in the document pointed by doc_path.
        doc_path is an instance of mglib.path.DocumentPath

        In case of success returns document's new version.

        new_order is a list of following format:

            [
                {'page_num': 2, page_order: 1},
                {'page_num': 1, page_order: 2},
                {'page_num': 3, page_order: 3},
                {'page_num': 4, page_order: 4},
            ]
        Example above means that in current document of 4 pages,
        first page was swapped with second one.
        page_num    = older page order
        page_order  = current page order
        So in human language, each hash is read:
            <page_num> now should be <page_order>
        """
        src_doc_path = doc_path
        dst_doc_path = DocumentPath.copy_from(
            src_doc_path,
            version=doc_path.version + 1
        )
        self.make_sure_path_exists(
            self.abspath(dst_doc_path)
        )

        stapler.reorder_pages(
            src=self.abspath(src_doc_path),
            dst=self.abspath(dst_doc_path),
            new_order=new_order
        )

        page_count = self.get_pagecount(doc_path)

        if len(new_order) > page_count:
            logger.error(
                f"deleted_pages({new_order}) > page_count({page_count})"
            )
            return

        for item in new_order:
            for step in Steps():
                src_page_path = PagePath(
                    document_path=src_doc_path,
                    page_num=int(item['page_num']),
                    step=step,
                    page_count=len(new_order)
                )
                dst_page_path = PagePath(
                    document_path=dst_doc_path,
                    page_num=int(item['page_order']),
                    step=step,
                    page_count=len(new_order)
                )
                self.copy_page(
                    src_page_path=src_page_path,
                    dst_page_path=dst_page_path
                )

        return doc_path.version + 1

    def delete_pages(
        self,
        doc_path,
        page_numbers,
        skip_migration=False
    ):
        """
        Delets pages in the document pointed by doc_path.
        doc_path is an instance of mglib.path.DocumentPath

        In case of success returns document's new version.
        """

        if not isinstance(page_numbers, list):
            logger.error("Expecting list argument")
            return False

        src_doc_path = doc_path
        dst_doc_path = DocumentPath.copy_from(
            src_doc_path,
            version=doc_path.version + 1
        )
        self.make_sure_path_exists(
            self.abspath(dst_doc_path)
        )
        stapler.delete_pages(
            self.abspath(src_doc_path),
            self.abspath(dst_doc_path),
            page_numbers
        )

        if skip_migration:
            return doc_path.version + 1

        page_count = self.get_pagecount(doc_path)

        if len(page_numbers) > page_count:
            logger.error(
                f"deleted_pages({page_numbers}) > page_count({page_count})"
            )
            return

        assigns = get_assigns_after_delete(
            total_pages=page_count,
            deleted_pages=page_numbers
        )
        for a in assigns:
            for step in Steps():
                src_page_path = PagePath(
                    document_path=src_doc_path,
                    page_num=a[1],
                    step=step,
                    page_count=page_count
                )
                dst_page_path = PagePath(
                    document_path=dst_doc_path,
                    page_num=a[0],
                    step=step,
                    page_count=page_count - len(page_numbers)
                )
                self.copy_page(
                    src_page_path=src_page_path,
                    dst_page_path=dst_page_path
                )

        return doc_path.version + 1

    def paste_pages(
        self,
        dest_doc_path,
        data_list,
        dest_doc_is_new=False,
        after_page_number=False,
        before_page_number=False
    ):
        """
        Pastes pages in the document pointed by dest_doc_path
        from src_doc_path. Both dest and src are instances of
        mglib.path.DocumentPath
        """
        next_ver_dp = DocumentPath.copy_from(
            dest_doc_path,
            version=dest_doc_path.version + 1
        )
        self.make_sure_path_exists(
            self.abspath(next_ver_dp)
        )

        stapler.paste_pages(
            src=self.abspath(dest_doc_path),
            dst=self.abspath(next_ver_dp),
            data_list=data_list,
            dst_doc_is_new=dest_doc_is_new,
            after_page_number=after_page_number,
            before_page_number=before_page_number
        )

        if not dest_doc_is_new:
            # migrate document's own pages from previous
            # version (this differs from pasting into newly
            # created docs)
            pcount = self.get_pagecount(dest_doc_path)
            data_list.insert(
                0,
                {
                    'doc_path': dest_doc_path,
                    'page_nums': list(range(1, pcount + 1))
                }
            )

        dest_page_num = 1
        dest_page_count = sum([
            len(item['page_nums']) for item in data_list
        ])
        for item in data_list:
            src_path = item['doc_path']
            for page_num in item['page_nums']:
                for step in Steps():
                    src_page_path = PagePath(
                        document_path=src_path,
                        page_num=int(page_num),
                        step=step,
                        page_count=self.get_pagecount(src_path)
                    )
                    dst_page_path = PagePath(
                        document_path=next_ver_dp,
                        page_num=dest_page_num,
                        step=step,
                        page_count=dest_page_count
                    )
                    logger.debug(f"src={src_page_path}  dst={dst_page_path}")
                    self.copy_page(
                        src_page_path=src_page_path,
                        dst_page_path=dst_page_path
                    )
                dest_page_num += 1

        return dest_doc_path.version + 1


class FileSystemStorage(Storage):
    pass



import os
import logging
import importlib.machinery
import importlib.util


logger = logging.getLogger(__name__)

SAFE_EXTENSIONS = [
    '.txt',
    '.jpg',
    '.jpeg',
    '.png',
    '.hocr',
    '.pdf',
    '.tiff'
]


def get_bool(key, default="NO"):
    """
    Returns True if environment variable named KEY is one of
    "yes", "y", "t", "true" (lowercase of uppercase)

    otherwise returns False
    """
    env_var_value = os.getenv(key, default).lower()
    YES_VALUES = ("yes", "y", "1", "t", "true")
    if env_var_value in YES_VALUES:
        return True

    return False


def safe_to_delete(place):
    if not os.path.exists(place):
        logging.warning(
            f"Trying to delete not exising folder"
            f" {place}"
        )
        return False

    for root, dirs, files in os.walk(place):
        for name in files:
            base, ext = os.path.splitext(name)
            if ext.lower() not in SAFE_EXTENSIONS:
                logger.warning(
                    f"Trying to delete unsefe location: "
                    f"extention={ext} not found in {SAFE_EXTENSIONS}"
                )
                return False

    return True


def get_assigns_after_delete(total_pages, deleted_pages):
    """
    given total pages and a list of deleted pages - returns
    a list of assignations of pages:
        [new_version_page_num, old_version_page_num]
    Example 1:
    total_pages: 6
    deleted_pages: [1, 2]
    returns: [
        [(1, 3),  (2, 4), (3, 5), (4, 6)]
        # page #1 gets info from prev page #3
        # page #2 ... #4
        ...
        # page #4 ... #6
    ]

    Example 2:
    total pages: 5
    deleted_pages [1, 5]
    returns: [
        [(1, 2), (2, 3), (3, 4)
    ]

    Example 3:
    total pages: 5
    deleted_pages [2, 3]
    returns: [
        [(1, 1), (2, 4), (3, 5)
        # page #1 stays unaffected
        # page #2 gets the info from page number 4
        # page #3 gets info from page #5
    ]
    """
    if total_pages < len(deleted_pages):
        err_msg = "total_pages < deleted_pages"
        raise ValueError(err_msg)

    # only numbers of pages which were not deleted
    pages = [
        page for page in list(range(1, total_pages + 1))
        if page not in deleted_pages
    ]

    page_numbers = range(1, len(pages) + 1)

    return list(zip(page_numbers, pages))


def load_config(config_file):

    loader_ = importlib.machinery.SourceFileLoader(
        "config",
        config_file
    )
    spec = importlib.util.spec_from_file_location(
        "config", config_file, loader=loader_
    )
    mod = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(mod)
    # and stop looking for ther configs.
    cfg_papermerge = vars(mod)
    cfg_file_found = True

    return cfg_papermerge, cfg_file_found


def try_load_config(config_locations, config_env_var_name):

    cfg_file_found = False
    cfg_papermerge = {}

    for config_file in config_locations:
        if os.path.exists(config_file):
            cfg_papermerge, cfg_file_found = load_config(config_file)
            if cfg_file_found:
                break

    if not cfg_file_found:
        config_file = os.environ.get(config_env_var_name, False)
        if config_file:
            try:
                cfg_papermerge, cfg_file_found = load_config(config_file)
            except FileNotFoundError:
                err_msg = "Failed attempted to read" +\
                    f" configuration file '{config_file}'" +\
                    f" pointed by environment variable '{config_env_var_name}'"
                raise FileNotFoundError(err_msg)

    return cfg_papermerge



import os
import unittest

from mglib.pdfinfo import get_pagecount

BASE_DIR = os.path.dirname(
    os.path.abspath(__file__)
)

DATA_DIR = os.path.join(
    BASE_DIR, "data"
)


def get_filepath(filename):
    return os.path.join(DATA_DIR, filename)


class TestPDFinfo(unittest.TestCase):

    def test_basic_pdf(self):
        page_count = get_pagecount(get_filepath("berlin.pdf"))

        self.assertEqual(
            page_count,
            2
        )

    def test_basic_jpeg(self):
        page_count = get_pagecount(get_filepath("berlin.jpeg"))

        self.assertEqual(
            page_count,
            1
        )

    def test_basic_jpg(self):
        page_count = get_pagecount(get_filepath("berlin.jpg"))

        self.assertEqual(
            page_count,
            1
        )

    def test_basic_png(self):
        page_count = get_pagecount(get_filepath("berlin.png"))

        self.assertEqual(
            page_count,
            1
        )

    def test_basic_tiff(self):
        # in case input file has extention tiff extension
        # it will internally call get_tiff_pagecount method
        page_count = get_pagecount(get_filepath("text.tiff"))

        self.assertEqual(
            page_count,
            2
        )



import logging
import subprocess


logger = logging.getLogger(__name__)


def run(cmd):
    logger.debug(
        f"Run:{'|'.join(cmd)}"
    )

    ret = subprocess.run(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        encoding="utf-8"
    )

    if ret.returncode != 0:
        logger.error((
            f"returncode={ret.returncode}"
            f" stdout={ret.stdout}"
            f" stderr={ret.stderr}"
        ))



import os
from pathlib import Path
import unittest
from mglib import mime

DATA_DIR = os.path.join(
    Path(__file__).parent,
    'data'
)


class TestConvert(unittest.TestCase):

    def test_basic_command_call(self):

        file_path = os.path.join(
            DATA_DIR,
            "berlin.pdf"
        )
        mime_type = mime.Mime(filepath=file_path)
        self.assertTrue(
            mime_type.is_pdf()
        )



import unittest
from mglib.step import Step


class TestStep(unittest.TestCase):

    def test_step(self):
        step = Step(1)
        self.assertFalse(
            step.is_thumbnail,
            f"{step} is is_thumbnail, but it should not be!"
        )



import os
import shutil


class TemporaryNode:
    """
    Handy class when it comes to testing files/directories
    structures.

    Example of usage:

    with TemporaryNode(MEDIA_ROOT) as media_root:
        docs = media_root.add_folder("docs")
        res = media_root.add_folder("results")
        f1 = docs.add_folder("user_1/document_2")
        f1.add_file("doku.pdf")
        res.add_folder("user_1/document_2/pages")

        docp = DocumentPath(
            user_id=1,
            document_id=2,
            file_name='doku.pdf'
        )

        storage.delete_document(docp)

        self.assertFalse(
            docs.exists()
        )

        self.assertFalse(
            res.exists()
        )
    """

    def __init__(self, location):
        self._location = location

    @property
    def location(self):
        return self._location

    def __enter__(self):
        if not os.path.exists(self.location):
            os.makedirs(self.location, exist_ok=True)

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if os.path.exists(self.location):
            if os.path.isdir(self.location):
                shutil.rmtree(self.location)
            else:
                os.remove(self.location)

    def add_folder(self, folder):
        new_location = os.path.join(
            self.location,
            folder
        )
        os.makedirs(new_location, exist_ok=True)
        return TemporaryNode(new_location)

    def exists(self):
        return os.path.exists(self.location)

    def add_file(self, file):
        return self



class Step:
    # Q: What is ``Step`` and why it was a bad decision to introduce it?
    #
    # A: ``Step`` class is closely related to zooming in/zooming out
    # a specific page in the document in the frontend (javascript code).
    #
    # When user opens the document in document viewer, he/she actually
    # sees an image with text over it (text overlay). Text overlay is
    # created from hocr data. Very important point here, is that
    # text hocr data corresponds to (extracted, format jpeg) image of the page
    # of VERY SAME width/height. Again, hocr file and respective image file
    # of the page MUST HAVE SAME WIDTH AND HEIGHT.
    #
    # Each step is meant to be a specific zoom value of the page. Thus, step
    # 2, which corresonds to LIST[2] % = 75 % of the page initial logical size
    # of WIDTH_100p = 1240.
    # When user zooms in/zooms out - a new hocr file is downloaded
    # corresponding to that zoom step. As you may guess, user can zoom only
    # 125%, 100%, 75% and 50%. Value of 10% corresponds to thumbnail of the
    # document and does not count as 'real' step.
    #
    # Instead of doing this step thingy, it would have been better to drop
    # the entire step concept. Much better solution for zoom in/zoom out would
    # have been to download one SVG file for each page (instead of hocr) and
    # SVG file of respective page should contain embedded image
    # (binary jpeg; yes SVG format allows embedding of binary formats!) and
    # correctly mapped text overlay (built from hocr file). User later
    #  can zoom in/zoom out using SVG transforations in frontend!
    #
    # The good things about SVG solutions are:
    #
    # * there will be 4X less OCR required (corresponding to
    #     hOCR of each step minus thumbnail/10% step)
    # * will simplify front-end code as SVG (= hocr + jpeg) will be
    #       generated on the on server side
    # * eliminate conept of Step entirely
    #    (there will be only one SVG file per page)
    # * increase front-end and back-end performance as only one file SVG file
    #   will be sent back and forth (from backend to frontend)
    #
    # width of a document when displayed as 100%.
    WIDTH_100p = 1240
    PERCENT = 100
    LIST = [125, 100, 75, 50, 10]

    # aspect ration for A4 paper is h = w * 1.41
    # for 100
    # 100 => w = 1240, h = 1748
    # 50  => w = 620, h = 874

    def __init__(self, current=1):
        self.current = current

    @property
    def width(self):
        p = self.percent / 100
        return int(p * Step.WIDTH_100p)

    @property
    def is_thumbnail(self):
        return self.percent < 50

    @property
    def is_for_hocr(self):
        return not self.is_thumbnail

    @property
    def percent(self):
        return Step.LIST[self.current]

    def __str__(self):
        return f"Step(percent={self.percent}, width={self.width})"

    def __repr__(self):
        return self.__str__()


class Steps:
    def __init__(self):
        self.steps = [Step(0), Step(1), Step(2), Step(3), Step(4)]

    def __iter__(self):
        return iter(self.steps)
